#/**********************************************************************************
#* Copyright (C) 2017-2020 by KEYARK, INC. - All Rights Reserved.
#* All information contained herein is proprietary and confidential to KEYARK, INC.
#* Any use, reproduction, or disclosure without the written
#* permission of KEYARK, INC. is prohibited.
#* Written by Natalia Belyaeva <nbelyaeva@keyark.com>
#***********************************************************************************/

from enum import Enum
from typing import List

# types: number, char, date, multiple, composition
class KEYOBJECT_TYPE(Enum):
    """Keyobjects types:
        NUMBER = 0
        CHAR = 1
        DATE = 2
        INTEGER = 3
        MULTIPLE = 4
        COMPOSITION = 5
    """
    NUMBER = 0
    CHAR = 1
    DATE = 2
    INTEGER = 3
    MULTIPLE = 4
    COMPOSITION = 5


class KeyObject(object):
    """
    KeyObject base class
    """
    #__name__ = None
    def __init__(self, name: str = None):
        self.__name__ = name

    def keyobject_type(self):
        pass

    def get_name(self) -> str:
        """
        :return: str - Keyobject name
        """
        return self.__name__

    def get_children(self) -> List:
        """Children of the Keyobject class
        :return: List
        """
        return []

    def serialize(self):
        pass

    """
    def rename(self, name : str, schema=None) -> KeySqlStatement:
        old_name = self.__name__
        self.__name__ = name
        return KeySqlRenameStoreStatement(old_name, self.__name__, schema=schema)
    """


class KeyObjectNumber(KeyObject):
    def __init__(self, name: str = None):
        """Create KEyobject of type NUMBER
        :param name (str): name, should be nonempty and unique
        Example:
        KeyObjectNumber('kn')
        """
        KeyObject.__init__(self, name)

    def keyobject_type(self) -> KEYOBJECT_TYPE:
        """Type of Keyobject
        :return (KEYOBJECT_TYPE):
        Example:
            KeyObjectNumber('kn').keyobject_type()  ->  KEYOBJECT_TYPE.NUMBER
        """
        return KEYOBJECT_TYPE.NUMBER

    def keyobject_name(self) -> str:
        """Name of Keyobject
        :return (KEYOBJECT_NAME):
        Example:
            KeyObjectNumber('kn').keyobject_name()  ->  'kn'
        """
        return self.__name__

    def serialize(self) -> str:
        """Serialize Keyobject
        :return (str): str
        Example:
            KeyObjectNumber('kn').serialize()  ->  'kn NUMBER'
        """
        return  self.__name__ + ' ' +  KEYOBJECT_TYPE.NUMBER._name_


class KeyObjectInteger(KeyObject):
    def __init__(self, name: str = None):
        """Create Keyobject of type INTEGER
        :param name (str): name, should be nonempty and unique
        Example:
        KeyObjectInteger('kn')
        """
        KeyObject.__init__(self, name)

    def keyobject_type(self) -> KEYOBJECT_TYPE:
        """Type of Keyobject
        :return (KEYOBJECT_TYPE):
        Example:
            KeyObjectNumber('kn').keyobject_type()  ->  KEYOBJECT_TYPE.INTEGER
        """
        return KEYOBJECT_TYPE.INTEGER

    def keyobject_name(self) -> str:
        """Name of Keyobject
        :return (KEYOBJECT_NAME):
        Example:
            KeyObjectNumber('kn').keyobject_name()  ->  'kn'
        """
        return self.__name__

    def serialize(self) -> str:
        """Serialize Keyobject
        :return (str): str
        Example:
            KeyObjectNumber('kn').serialize()  ->  'kn NUMBER'
        """
        return  self.__name__ + ' ' +  KEYOBJECT_TYPE.INTEGER._name_


class KeyObjectChar(KeyObject):
    def __init__(self, name: str = None):
        """Create KEyobject of type CHAR
        :param name (str): name, should be nonempty and unique
        Example:
        KeyObjectChar('kc')
        """
        KeyObject.__init__(self, name)

    def keyobject_type(self) -> KEYOBJECT_TYPE:
        """Type of Keyobject
        :return (KEYOBJECT_TYPE):
        Example:
            KeyObjectChar('kc').keyobject_type()  -> KEYOBJECT_TYPE.CHAR
        """
        return KEYOBJECT_TYPE.CHAR

    def serialize(self) -> str:
        """Serialize Keyobject
        :return (str): str
            KeyObjectChar('kc').serialize()  -> 'kc CHAR'
        """
        return self.__name__ + ' ' +  KEYOBJECT_TYPE.CHAR._name_


class KeyObjectDate(KeyObject):
    def __init__(self, name: str):
        """Create KEyobject of type DATE
        :param name (str): name, should be nonempty and unique
        Example:
        KeyObjectDate('kd')
        """

        KeyObject.__init__(self, name)

    def keyobject_type(self) -> KEYOBJECT_TYPE:
        """Type of Keyobject
        :return (KEYOBJECT_TYPE):
        Example:
            KeyObjectDate('kd').keyobject_type() -> KEYOBJECT_TYPE.DATE
        """
        return KEYOBJECT_TYPE.DATE

    def serialize(self) -> str:
        """Serialize Keyobject
        :return (str): str
        Example:
            KeyObjectDate('kd').serialize() -> 'kd DATE'
        """
        return self.__name__ + ' ' +  KEYOBJECT_TYPE.DATE._name_

class KeyObjectMultiple(KeyObject):
    #ref  #keyobject
    def __init__(self, name: str, ref: KeyObject):
        """Create KEyobject of type MULTIPLE
        :param name (str): name, should be nonempty and unique
        Example:
        KeyObjectMultiple('km', KeyObjectNumber('kn'))
        """
        KeyObject.__init__(self, name)
        self.__ref__ = ref

    def keyobject_type(self) -> KEYOBJECT_TYPE:
        """Type of Keyobject
        :return (KEYOBJECT_TYPE):
        Example:
            KeyObjectMultiple('km', KeyObjectNumber('kn')).keyobject_type() -> KEYOBJECT_TYPE.MULTIPLE
        """
        return KEYOBJECT_TYPE.MULTIPLE

    def get_children(self) -> List[KeyObject]:
        """Get Children
        :return (List[KeyObject]): list of KeyObjects
        Example:
        children = KeyObjectMultiple('km', KeyObjectNumber('kn')).get_children()
        res = [child.serialize() for child in children]   -> ['kn NUMBER']
        Children for KeyObjectMultiple are always an array of length 1
        (to be of the same type as for KeyObjectComposition)
        """
        return [self.__ref__] #  [self.__ref__]

    def serialize(self) -> str:
        """Serialize Keyobject
        :return (str): str
        Example:
            KeyObjectMultiple('km', KeyObjectNumber('kn')).serialize() -> 'km{kn MULTIPLE}'
        """
        #return self.__name__  + '{' + self.__ref__[0].__name__ + ' MULTIPLE' + '}'

        return self.__name__  + '{' + self.__ref__.__name__ + ' MULTIPLE' + '}'

        #return self.get_name()  + '{' + self.__ref__.get_name() + ' MULTIPLE' + '}'

class KeyObjectComposition(KeyObject):
    #childen  #keyobject list/dic (name: keyobject)
    __name =None
    __children__ = []

    def __init__(self, name: str, children : List[KeyObject]):
        """Create KEyobject of type COMPOSITION
        :param name (str): name, should be nonempty and unique
        :param children (List[KeyObject]): children
        Example:
        KeyObjectMultiple('km', KeyObjectNumber('kn'))
        """

        KeyObject.__init__(self, name)
        self.__children__ = children

    def keyobject_type(self) -> KEYOBJECT_TYPE:
        """Type of Keyobject
        :return (KEYOBJECT_TYPE):
        Example:
            kcomp = KeyObjectComposition('kcomp', [KeyObjectNumber('kn'), KeyObjectChar('kc')])
            kcomp.keyobject_type() -> KEYOBJECT_TYPE.COMPOSITION
        """
        return KEYOBJECT_TYPE.COMPOSITION

    def get_children(self) -> List[KeyObject]:
        """Get Children
        :return (List[KeyObject]): list of KeyObjects
        Example:
        children = KeyObjectComposition('kcomp', [KeyObjectNumber('kn'), KeyObjectChar('kc')]).get_children()
        res = [child.serialize() for child in children]  -> ['kn NUMBER', 'kc CHAR']
        """
        return self.__children__

    def serialize(self) -> str:
        """Serialize Keyobject
        :return (str): str
        Example:
            children = KeyObjectComposition('kcomp', [KeyObjectNumber('kn'), KeyObjectChar('kc')]).get_children()
            res = [child.serialize() for child in children]  -> ['kn NUMBER', 'kc CHAR']
        """
        ser_str_arr = []
        for mem in self.__children__:
            ser_str_arr.append(mem.__name__)
        return self.__name__ + '{' + ','.join(ser_str_arr) + '}'


class KeyObjectBuilder():

    @staticmethod
    def create_keysql_object_from_string(sss) -> KeyObject:
        if "{" not in sss:
            sss_parsed = sss.split(' ')
            if sss_parsed[1] == "CHAR":
                return KeyObjectChar(sss_parsed[0])
            elif sss_parsed[1] == "NUMBER":
                return KeyObjectNumber(sss_parsed[0])
            elif sss_parsed[1] == "DATE":
                return KeyObjectDate(sss_parsed[0])
        else:
            start = sss.find('{')
            end = sss.find('}')
            childs_names = sss[start+1:end].strip().split(',')
            parent = sss[:start].strip()
            if len(childs_names) == 1 and len(childs_names[0].split(' ')) == 2\
                and childs_names[0].split(' ')[1] == 'MULTIPLE':
                return KeyObjectMultiple(parent, KeyObject(childs_names[0].split(' ')[0]))
            else:
                childs = []
                for child in childs_names:
                    childs.append(KeyObject(child))
                return KeyObjectComposition(parent, childs)

        return

'''
class keyobject_char(keyobject)
  pass
class keyobject_date(keyobject)::
  pass
class keyobject_multiple(keyobject):
  ref  #keyobject
class keyobject_composition(keyobject):
  childs  #keyobject list/dic (name: keyobject)
create:
  create keyobject <name> <type> in catalog cname;
  1. create keyobject nn number in catalog xyz
  2. create keyobject cc char in catalog xyz
  3. create keyobject dd date in catalog xyz
  4 create keyobject mm{x multiple} in catalog xyz
  5 create keyobject cc{x,y,z} in catalog xyz
drop:
  drop keyobject <name> from catalog cname;
alter
   alter keyobject <name> = <new_name> <new  type> in catalog cname
show keyobject <name> from catalog cname
'''
if __name__ == "__main__":
    """
    Serialization example
    KEYOBJECT_TYPE.NUMBER
    kn output:
     kn NUMBER
    multiple km output:
    km{km1 MULTIPLE}
     composition kkk:
    kkk{kn,km}
    """
    """
    Examples:
    {BALL {COLOR,NUM},
    BALLPACK {BALLSET,BRAND,NAME,UPC},
    BALLS {BALL MULTIPLE},
    CITY CHAR,
    COLOR CHAR,
    DIAM_MM NUMBER,
    """

    obj = KeyObjectBuilder().create_keysql_object_from_string("BALLS {BALL MULTIPLE}")
    assert obj.get_name() == 'BALLS'
    obj = KeyObjectBuilder().create_keysql_object_from_string("BALL {COLOR,NUM}")
    assert obj.get_name() == 'BALL'
    assert obj.keyobject_type() == KEYOBJECT_TYPE.COMPOSITION
    for x in obj.get_children():
         assert x.get_name() in ['COLOR','NUM']
    #kn = KeyObjectNumber('kn')
    assert KeyObjectNumber('kn').serialize() == "kn NUMBER"
    assert KeyObjectNumber('kn').keyobject_type() == KEYOBJECT_TYPE.NUMBER

    #kc = KeyObjectChar('kc')
    assert KeyObjectChar('kc').serialize() == "kc CHAR"
    assert KeyObjectChar('kc').keyobject_type() == KEYOBJECT_TYPE.CHAR

    #kd = KeyObjectDate('kd')
    assert KeyObjectDate('kd').serialize() == "kd DATE"
    assert KeyObjectDate('kd').keyobject_type() == KEYOBJECT_TYPE.DATE

    #kd = KeyObjectInteger('kd')
    assert KeyObjectInteger('kd').serialize() == "kd INTEGER"
    assert KeyObjectInteger('kd').keyobject_type() == KEYOBJECT_TYPE.INTEGER

    #km = KeyObjectMultiple('km', KeyObjectNumber('kn'))
    assert KeyObjectMultiple('km', KeyObjectNumber('kn')).serialize() == "km{kn MULTIPLE}"
    assert KeyObjectMultiple('km', KeyObjectNumber('kn')).keyobject_type() == KEYOBJECT_TYPE.MULTIPLE

    children = KeyObjectMultiple('km', KeyObjectNumber('kn')).get_children()
    assert len(children) == 1
    assert children[0].serialize() == "kn NUMBER"

    children = KeyObjectMultiple('km', KeyObjectNumber('kn')).get_children()
    assert [child.serialize() for child in children] == ['kn NUMBER']

    #kcomp = KeyObjectComposition('kcomp', [KeyObjectNumber('kn'), KeyObjectChar('kc')])
    assert KeyObjectComposition('kcomp', [KeyObjectNumber('kn'), KeyObjectChar('kc')]).serialize() == "kcomp{kn,kc}"
    assert KeyObjectComposition('kcomp', [KeyObjectNumber('kn'), KeyObjectChar('kc')]).keyobject_type() == KEYOBJECT_TYPE.COMPOSITION

    kcomp = KeyObjectComposition('kcomp', [KeyObjectNumber('kn'), KeyObjectMultiple('km', KeyObjectNumber('kn'))])
    assert KeyObjectComposition('kcomp', [KeyObjectNumber('kn'), KeyObjectMultiple('km', KeyObjectNumber('kn'))]).serialize() == "kcomp{kn,km}"
    assert KeyObjectComposition('kcomp', [KeyObjectNumber('kn'), KeyObjectMultiple('km', KeyObjectNumber('kn'))]).keyobject_type() == KEYOBJECT_TYPE.COMPOSITION

    children = KeyObjectComposition('kcomp', [KeyObjectNumber('kn'), KeyObjectChar('kc')]).get_children()
    assert sorted([child.serialize() for child in children]) == ['kc CHAR', 'kn NUMBER']
