"""
# /**********************************************************************************
#* Copyright (C) 2017-2020 by KEYARK, INC. - All Rights Reserved.
#* All information contained herein is proprietary and confidential to KEYARK, INC.
#* Any use, reproduction, or disclosure without the written
#* permission of KEYARK, INC. is prohibited.
#* Written by Natalia Belyaeva <nbelyaeva@keyark.com>
#***********************************************************************************/
"""

import locale
locale.setlocale(locale.LC_ALL, 'en_US.utf-8')
import time

from tornado.ioloop import IOLoop
from tornado.gen import convert_yielded

from keysql_client.keysql_connector_async import KeySqlConnectorAsync, KeySqlConnectException
from keysql_client.keysql_statement import KeySqlStatement
from keysql_client.keysql_exception import KeySqlExceptionMessage, KeySqlException
from keysql_client.keysql_reply_statement import KeySqlReplyStatement

from keysql_client.keysql_parser_query_parameterized import KeySqlParserQueryParameterized

from typing import List

import logging

import tracemalloc

class KeySqlConnectionAsync:
    """Connection with KeySql server
    """
    last_str_result = ""
    cum_execute_time_in_sec =  0.0
    TOTAL_LEN_BUG_REPORT = 400

    def __init__(self, host : str, port : int, ioloop : IOLoop = IOLoop.current()):
        self._connector = KeySqlConnectorAsync(host,port, ioloop)
        self._reply_header = None

    async def connect_user(self, user, password, ssl: bool = True):
        reply = None
        if ssl:
            reply = await self._connector.ssl_connect_async(user, password)
        else:
            reply = await self._connector.connect_user(user, password)

        self._connected = reply
        return reply


    def __exit__(self):
        self._connector._disconnect()

    def disconnect(self):
        """Disconnect with KeySql server
        :return: None
        """
        self.__exit__()

    def is_connected(self):
        """Check if connected with KeySql server
        :return: True or False
        """
        return self._connector.is_connected()

    def get_header_reply(self):
        """Get header
        :return: Dictionary of attributes, header['exit_code'] = 0 means successful query
        """
        return self._reply_header

    def get_last_str_result(self):
        """Get last query result in a string format
        :return: string
        """
        return self.last_str_result

    async def execute(self, statement, *args, base_classes: List=[], keep_str_res=False, check_inj=False, format='keysql'):
        """Execute query on KeySql server
        :param statement (string, list of class:
        :param base_classes: List of classes to map (default: []  (works only now))
        :return: a list of classes
        :raise: Exception if query can not be performed (no connection, wrong syntax, etc.)
        """
        if type(statement) == str:
            statement = statement.strip()
            if len(statement) < 1:
                raise KeySqlConnectException(KeySqlExceptionMessage.EMPTY_STATEMENT)

            if statement[-1] == ';':
                #all_statements = statement[:-1]  # statement[:-1].split(';')
                statement = statement[:-1]
            if ';' in statement:
                statements = self.extract_statements(statement)
            else:
                statements = [statement]

            if len(statements) > 1 and len(args) != 0:
                raise KeySqlException(-1, KeySqlExceptionMessage.PARAMETERIZED_ONE_QUERY_ONLY)
            elif len(statements) == 1 and len(args) > 0:
                statements[0] = KeySqlParserQueryParameterized.process_input_query(statements[0], *args)

            # no arguments now
            #else:  # if len(statements) > 1:
            #return self.execute(statements, base_classes=base_classes, keep_str_res=keep_str_res,
            #                       check_inj=check_inj)

                #return self.execute(statements[0], base_classes=base_classes, keep_str_res=keep_str_res,
                #
                #                     check_inj=check_inj)
                #self._reply_header, reply = self._connector.send_resv(statements[0])
            #else:
            #    return self.execute(statements, base_classes=base_classes, keep_str_res=keep_str_res, check_inj=check_inj)

                # return only the result of last statement
                #for statement_i in all_statements:
                #    reply = self.execute(statement_i, base_classes)
                #return reply

            # statements = self.extract_statements(statement)
            for statement in statements:
                logging.debug('---> final_request1: ' + str(statements))

                self.cum_execute_time_in_sec = time.time()
                self._reply_header, reply = await self._connector.send_resv(statement, format)
                self.cum_execute_time_in_sec = time.time() - self.cum_execute_time_in_sec

                #self._reply_header, reply = self._connector.send_resv(str.encode(statement, encoding="utf-8"))
                if keep_str_res:
                    self.last_str_result = reply
                logging.debug('time: ' + str(self._reply_header["time"]) + '---> reply' + reply)
                if self._reply_header['exit_code'] != 0:
                    raise KeySqlException(self._reply_header['exit_code'], reply)
                if format == 'keysql':
                    reply = KeySqlReplyStatement(self._reply_header).get_keysql_statement().process_reply(reply,self._reply_header)
            return reply

        elif type(statement) == list:
            if len(statement) > 1 and len(args) > 0:
                raise KeySqlException(-1, KeySqlExceptionMessage.PARAMETERIZED_ONE_QUERY_ONLY)
            if len(statement) == 1:  #  and len(args) > 0:
                reply = await self.execute(statement[0], *args, base_classes=base_classes, keep_str_res=keep_str_res,
                                     check_inj=check_inj)
                return reply

            reply = ''
            # no arg here
            for st in statement:
                reply = await self.execute(st, base_classes=base_classes, keep_str_res=keep_str_res, check_inj=check_inj)
            return reply

        elif KeySqlStatement in statement.__class__.__mro__:   #  KeySqlStatement
            reply = '***no_reply***'
            try:
                final_request = statement.create_request()
                logging.debug('---> final_request2: ' + final_request[:self.TOTAL_LEN_BUG_REPORT] + "...")

                self.cum_execute_time_in_sec = time.time()
                self._reply_header, reply = await self._connector.send_resv(final_request, format)
                self.cum_execute_time_in_sec = time.time() - self.cum_execute_time_in_sec

                logging.debug('time: ' + str(self._reply_header["time"]) + '---> reply: ' + reply)
                if self._reply_header['exit_code'] == 0:
                    return statement.process_reply(reply, self._reply_header)
                else:
                    logging.warning('--- not success: ', self._reply_header['exit_code'])
                    logging.warning('---', reply)
                    # print_debug('---', self._reply_header['exit_code'], reply)
                    if self._reply_header is not None:
                        raise KeySqlException(self._reply_header['exit_code'], reply)
                    else:
                        raise KeySqlException(-1, reply)
            except ValueError as err:
                logging.exception("message")
                raise KeySqlException(-1, reply)
        else:
            logging.exception(KeySqlExceptionMessage.STATEMENT_TYPE_NOT_IMPLEMENTED)
            raise KeySqlException(-1, KeySqlExceptionMessage.STATEMENT_TYPE_NOT_IMPLEMENTED)

    def extract_statements(self, statement: str):
        stmts = []
        start = 0
        started_value = False
        for i, c in enumerate(statement):
            if c == "'":
                if not started_value:
                    started_value = True
                else:
                    started_value = False
            elif c == ";" and not started_value:  # not started_value and i > start + 1:
                stmts.append(statement[start:i].strip())
                start = i + 1
        if started_value:
            raise KeySqlException(-1, KeySqlExceptionMessage.IMPROPER_END_OF_STATEMENT)

        if start + 1 < len(statement):
            stmts.append(statement[start:len(statement)])
        return stmts

#-------------------------------------------------------------------------------
# for test async

async def conn1_async(my_local_host, my_port, my_user, my_pass, ioLoop):
    time_start = time.time()

    connection = KeySqlConnectionAsync(my_local_host, my_port, ioLoop)
    await connection.connect_user(my_user, my_pass)
    print('connection:', connection.is_connected())

    for ii in range(10):
        reply = None
        reply = await connection.execute('show catalogs; select * from world_bank_store')
        if reply:
            print("conn1_async: ", ii, '-'*50)
            #print(reply[:4], ' ...')
    print("1st finished in {} milisec.".format(1000 * (time.time() - time_start)))

    connection.disconnect()


async def conn2_async(my_local_host, my_port, my_user, my_pass, ioLoop):
    time_start = time.time()
    #connector2 = KeySqlConnectorAsyncNative(my_local_host, my_port, ioLoop)
    #await connector2.connect_user(my_user, my_pass)
    #print('connector2:', connector2.is_connected())
    connection = KeySqlConnectionAsync(my_local_host, my_port, ioLoop)
    await connection.connect_user(my_user, my_pass)
    print('connection:', connection.is_connected())

    for ii2 in range(10):
        reply = None
        reply = await connection.execute('select * from world_bank_store')
        if reply:
            print("conn2_async: ", ii2, '='*75)
            #print(reply[:38], ' ...')
    print("2nd finished in {} milisec.".format(1000 * (time.time() - time_start)))

    connection.disconnect()

def callback(f):
    #print(f.result())
    pass

def two_conn(my_local_host, my_port, my_user, my_pass, ioLoop):
    print('time start: ', time.time())
    f1 = convert_yielded(conn1_async(my_local_host, my_port, my_user, my_pass, ioLoop))
    ioLoop.add_future(f1, callback)
    f2 = convert_yielded(conn2_async(my_local_host, my_port, my_user, my_pass, ioLoop))
    ioLoop.add_future(f2, callback)
    print('time end  : ', time.time())

if __name__ == "__main__":
    pass
