"""
#/**********************************************************************************
#* Copyright (C) 2017-2020 by KEYARK, INC. - All Rights Reserved.
#* All information contained herein is proprietary and confidential to KEYARK, INC.
#* Any use, reproduction, or disclosure without the written
#* permission of KEYARK, INC. is prohibited.
#* Written by Natalia Belyaeva <nbelyaeva@keyark.com>
#***********************************************************************************/
"""
import logging
from keysql_client.keysql_statement import KeySqlStatement
from keysql_client.keysql_exception import KeySqlException, KeySqlExceptionMessage
from keysql_client.keysql_mapper import KeySqlMapper
from keysql_client.keysql_parser import KeySqlParser
from keysql_client.keysql_parser_packed_typed_format import KeySqlParserPackedFormat
import json

class SelectStatement(KeySqlStatement):
    """ Select Statement
    """
    _select = None
    _arg_list = []
    _store_name = None
    _store_alias = None
    _join_store_name = None
    _join_store_alias = None
    _join_on = None
    _where = None
    _group_by = None
    _having = None
    _order_by = None
    _limit = None
    _into = None
    _tail = ""
    _to_dict = False

    _header = None

    def __init__(self) -> None:
        self._select = None
        self._arg_list = []
        self._store_name = None
        self._store_alias = None
        self._join_store_name = None
        self._join_store_alias = None
        self._join_on = None
        self._where = None
        self._group_by = None
        self._having = None
        self._order_by = None
        self._limit = None
        self._into = None
        self._tail = ""
        self._to_dict = False


    def set_default_values(self):
        self.__init__()


    #TODO
    # for any type of objects, strings for now
    def select(self, *args):
        """
        :param args (str, object names, dictionary , or list of those):
        :return: self
        """
        #print('total args: ', len(*args))
        arg_list = []
        for arg in args:
            if type(arg) == str:
                arg_list.append(arg)
            elif type(arg) == dict:
                arg_list.append(self.make_dict_arg(arg))
            else:
                arg_list.append(arg.__class__.__name__)
            self._arg_list.append(arg)
        self._select = 'select {}'.format(','.join(arg_list))
        return self

    def make_dict_arg(self,arg):
        if type(arg) == str:
            return arg
        r = ""
        if type(arg) == dict:
            for key,value in arg.items():
                if r != "":
                    r += ","
                r += self.make_dict_arg(value)
                r += " AS "
                r += str(key)
        elif type(arg) == list:
            r = "{"
            for x in arg:
                if r != "{":
                    r += ","
                r += x
            r += "}"
        return r


    def _from(self, store_name, alias=None):
        """From Clause
        :param store_name (str): store name
        :param alias (str):
        :return: self
        """
        self._store_name = store_name  # ' from {}'.format(store_name)
        self._store_alias = alias
        return self

    def join(self, store_name, on_condition, alias=None):
        """Join clause
        :param store_name (str): store name
        :param on_condition (str): condition as string
        :param alias (str): alias of results
        :return: self
        """
        self._join_store_name = store_name
        self._join_store_alias = alias
        if self._join_store_alias is None:
            self._join_on = ' join {} on {}'.format(self._join_store_name, on_condition)
        else:
            self._join_on = ' join {} {} on {}'.format(self._join_store_name, self._join_store_alias, on_condition)
        self._tail += self._join_on
        return self

    def where(self, *args):
        """Where clause
        :param args (*str): variable number of strings(>= 1).
        :return: self
        Example:
        emails = conn.execute(store.select("last_name","email").where(":first_name = 'John'"))
         :first_name means variable first_name
        """
        arg_list = []
        for arg in args:
            arg_list.append(str(arg))
        self._where = ' where {}'.format(','.join(arg_list))
        self._tail += self._where
        return self

    def group_by(self, *args):
        """Group by clause
        :param args (*str): variable number of strings(>= 1).
        :return: self
        Example:
        count = conn.execute(store.select({"last_name": "last_name", "cnt": "count(email)"}).group_by("last_name").having(":cnt > 1"))
        """
        arg_list = []
        for arg in args:
            if type(arg) == str:
                arg_list.append(arg)
            else:
                arg_list.append(arg.__class__.__name__)
        self._group_by = ' group by {}'.format(','.join(arg_list))
        self._tail += self._group_by
        return self

    def having(self, *args):
        """Having Clause
        :param args (*str):  variable number of strings(>= 1).
        :return: self
        Example:
        count = conn.execute(store.select({"last_name": "last_name", "cnt": "count(email)"}).group_by("last_name").having(":cnt > 1"))
        """
        arg_list = []
        for arg in args:
            arg_list.append(str(arg))
        self._having = ' having {}'.format(','.join(arg_list))
        self._tail += self._having
        return self

    def order_by(self, *args):
        """Order by Clause
        :param (*str):  variable number of strings(>= 1).
        :return: self
        Example:
        people = conn.execute(store.select(Person()).order_by("last_name")
        """
        arg_list = []
        for arg in args:
            arg_list.append(str(arg))
        self._order_by = ' order by {}'.format(','.join(arg_list))
        self._tail += self._order_by
        return self

    def limit(self, val):
        """Limit the number or results
        :param val (int): max number of return results
        :return (dict): d
        Example:
        people = conn.execute(store.select(Person()).order_by("last_name").limit(1)
        """
        self._limit = ' limit {}'.format(str(val))
        self._tail += self._limit
        return self

    def into_dict(self):
        """Transform results into dictionaries
        :return (dict): dictionaries
        Example:
        people = conn.execute(store.select(Person()).order_by("last_name").limit(1).into_dict())
        """
        self._to_dict = True
        return self

    #TODO
    #def into(self,var):
    #    self._into = ' into {}'.format(var)
    #    return self

    def create_request(self):
        #print('self._store_name: ', self._store_name)
        if self._store_name is None:
            raise KeySqlException(KeySqlExceptionMessage.STORE_IS_NOT_DEFINED)

        request = self._select + ' from ' + self._store_name

        if self._store_alias is not None:
            request += ' {}'.format(self._store_alias)

        request += self._tail
        logging.debug('select_request:' + request)
        return request

    def process_reply(self, reply, header=None, base_classes:list = []):

        logging.debug('reply before process:' + str(reply))
        format = 'keysql' if header is None else header['format']

        #print(reply, type(reply), format,  '-----------------------------------------try json')
        if header and header['format'] == 'json':
            if type(reply) is list:
                if reply == []:
                    return {}

                # TODO: remove, not needed
                """
                from keysql_client.keysql_instance import KeySql_Instance
                tmp = KeySql_Instance(KeySql_Instance(reply[0]))
                print(tmp)
                return [KeySqlMapper.get_default_class_from_dict(item, list(item.keys())[0]) for item in reply]
                """
                return reply

            if type(reply) is dict:
                return reply

            #print(reply, type(reply), '-----------------------------------------try json1')
            try:
                json_output = json.loads(reply)
                return json_output
            except KeySqlException(KeySqlExceptionMessage.WRONG_JSON_FORMAT) as ex:
                raise KeySqlException(self._reply_header['exit_code'], reply)

        # TODO: to remove obsolete
        elif header and header['format'] == 'keysql':
            instances = reply

            results = []
            for instance in instances:
                if self._to_dict:
                    curr = KeySqlMapper.keysql_inst_to_dict(instance)
                else:
                    if len(base_classes) == 0:
                        curr = KeySqlMapper.keysql_inst_to_default_class_only(instance)
                    else:
                        curr = KeySqlMapper.keysql_inst_to_class_inst(instance, base_classes)   #  ????? reply
                results.append(curr)

            return results

        elif header and header['format'] == 'packed':
            parser_packed_format = KeySqlParserPackedFormat()
            reply1 = parser_packed_format.process_select_reply(reply)
            return reply1

        else:
            raise KeySqlException(KeySqlExceptionMessage.FORMAT_KEYSQL_JSON_PACKED, header)

