#/**********************************************************************************
#* Copyright (C) 2017-2020 by KEYARK, INC. - All Rights Reserved.
#* All information contained herein is proprietary and confidential to KEYARK, INC.
#* Any use, reproduction, or disclosure without the written
#* permission of KEYARK, INC. is prohibited.
#* Written by Natalia Belyaeva <nbelyaeva@keyark.com>
#***********************************************************************************/
import logging
from typing import List, Optional
from keysql_client.keysql_statement import KeySqlStatement
from keysql_client.keysql_exception import KeySqlException, KeySqlExceptionMessage

class KeySqlCreateSequenceStatement(KeySqlStatement):
    """Show Sequence Statement
    String query equivalent:
    'create sequence <sequence_name> in catalog <catalog_name>'
    """
    _sequence_name = None
    _catalog_name = None

    def __init__(self, sequence_name: str=None, catalog_name: str = None):
        self._sequence_name = sequence_name
        self._catalog_name = catalog_name

    def create_request(self, header=None) -> str:
        request = 'create sequence {} in catalog {}'.format(self._sequence_name, self._catalog_name)
        logging.debug('create_sequence request: ' + request)
        return request


class KeySqlShowSequnceStatement(KeySqlStatement):
    """Show Sequence Statement
    String query equivalent:
    'show sequence <name> in catalog <catalog_name>'
    """
    _sequence_name = None
    _catalog_name = None

    def __init__(self, sequence_name: str = None, catalog_name: str = None) -> None:
        self._sequence_name = sequence_name
        self._catalog_name = catalog_name

    def create_request(self) -> str:
        return 'show sequence {} from catalog {}'.format(self._sequence_name, self._catalog_name)


class KeySqlShowSequncesStatement(KeySqlStatement):
    """Show Sequences Statement
    String query equivalent:
    'show sequences <name> from catalog <catalog_name>'
    """
    _catalog_name = None

    def __init__(self, catalog_name: str = None) -> None:
        self._catalog_name = catalog_name

    def create_request(self) -> str:
        return 'show sequences from catalog {}'.format(self._catalog_name)


class KeySqlListSequncesStatement(KeySqlStatement):
    """List Sequences Statement
    String query equivalent:
    'list sequences from catalog <catalog_name>'
    """
    _catalog_name = None

    def __init__(self, catalog_name: str = None) -> None:
        self._catalog_name = catalog_name

    def create_request(self) -> str:
        return 'list sequences from catalog {}'.format(self._catalog_name)


class KeySqlDescribeSequenceStatement(KeySqlStatement):
    """Describe Sequence Statement
    String query equivalent:
    'describe sequence <name> from catalog <catalog_name>'
    """
    _sequence_name = None
    _catalog_name = None

    def __init__(self, sequence_name: str = None, catalog_name: str = None) -> None:
        self._sequence_name = sequence_name
        self._catalog_name = catalog_name

    def create_request(self) -> str:
        return 'describe schema {}'.format(self._schema_name)


class KeySqlRenameSequenceStatement(KeySqlStatement):
    """Rename Sequence Statement
    String query equivalent:
    'alter sequence <sequence_name_from> from catalog  <catalog_name> with name = <sequence_name_to>'
    """
    _sequence_name_from = None
    _sequence_name_to = None
    _catalog_name = None

    def __init__(self, sequence_name_from: str = None, sequence_name_to: str = None, catalog_name: str = None) -> None:
        self._sequence_name_from = sequence_name_from
        self._sequence_name_to = sequence_name_to
        self._catalog_name = catalog_name

    def create_request(self) -> str:
        return 'alter sequence {} from catalog {} with name = {}'.format(self._sequence_name_from,
                                                                         self._catalog_name, self._sequence_name_to)


class KeySqlDropSequenceStatement(KeySqlStatement):
    """
    Drop Sequence Statement
    String query equivalent:
    'drop sequence <sequence_name> from catalog <catalog_name>'
    """

    def __init__(self, name: str = None, catalog_name = None) -> None:
        """
        :param name (str): sequence name to drop
        :param catalog_name (str): catalog name to drop
        Example:
        KeySqlDropSequenceStatement(<name>, <catalog_name>)
        """
        self._name = name
        self._catalog_name = catalog_name

    def create_request(self) -> str:
        return 'drop sequence {} from catalog {}'.format(self._name, self._catalog_name)


class KeySqlDropSequenceIfExistsStatement(KeySqlStatement):
    """
    Drop Sequence Statement
    String query equivalent:
    'drop sequence if exists <sequence_name> from catalog <catalog_name>'
    """

    def __init__(self, name: str = None, catalog_name = None) -> None:
        """
        :param name (str): sequence name to drop
        :param catalog_name (str): catalog name to drop
        Example:
       KeySqlDropSeuenceIfExistsStatement(<name>, <catalog_name>)
        """
        self._name = name
        self._catalog_name = catalog_name

    def create_request(self) -> str:
        return 'drop sequence if exists {} from catalog {}'.format(self._name,  self._catalog_name)


if __name__ == "__main__":

    create_schema = KeySqlCreateSequenceStatement('schema_name', 'catalog_name')
    drop_schema = KeySqlDropSequenceStatement('schema_name', 'catalog_name')
