"""
#/**********************************************************************************
#* Copyright (C) 2017-2020 by KEYARK, INC. - All Rights Reserved.
#* All information contained herein is proprietary and confidential to KEYARK, INC.
#* Any use, reproduction, or disclosure without the written
#* permission of KEYARK, INC. is prohibited.
#* Written by Natalia Belyaeva <nbelyaeva@keyark.com>
#***********************************************************************************/
"""
from typing import List, Optional
from keysql_client.keysql_store import  KeySqlStore
from keysql_client.keysql_catalog import KeySqlCatalog
from keysql_client.key_object import KeyObjectChar, KeyObjectInteger, KeyObjectNumber, KeyObjectMultiple, KeyObjectComposition
from keysql_client.key_object_statement import ShowKeyObjectStatement
from keysql_client.key_object_statement import ShowKeyObjectStatement
from keysql_client.keysql_statement_default import KeySqlDefaultStatement
from keysql_client.keysql_exception import KeySqlException, KeySqlExceptionMessage
from keysql_client.keysql_parser import KeySqlParser
import json
from collections import  OrderedDict

from abc import ABC, abstractmethod
class KeySqlShowStatement():
    def __init__(self):
        pass

    @abstractmethod
    def create_request(self):
        pass

    @abstractmethod
    def process_reply(self, reply, header):
        #return reply
        return header["instance_count"] if header else reply


class KeySqlShowStoreStatement(KeySqlShowStatement):
    """
    String query equivalent:
    show store <store_name>
    """
    _store = None
    #_catalog = None

    #def __init__(self, store_name: str = None, catalog_name: str = None) -> None:
    def __init__(self, store_name: str=None, schema_name: str = None) -> None:
        """ShowStoreStatement
        :param store_name (str): store name
        """
        self._store = store_name
        self._schema = schema_name


    def get_store_name(self) -> str:
        """
        store name
        :return: str - Keyobject name
        """
        return self._store


    def get_schema_name(self) -> str:
        """
        Schema name
        :return: str - schema name
        """
        return self._schema


    def create_request(self) -> str:
        #return 'show store {} for catalog {}'.format(self._store, self._catalog) # as it is uniq, catalog is not used
        if self._schema:
            return 'show store {}.{}'.format(self._schema, self._store)
        return 'show store {}'.format(self._store)

    def process_reply(self, reply: str, header=None) -> List[str]:
        res = KeySqlDefaultStatement().process_reply(reply)

        # example - keysql format
        # [{'store_info': {'store_name': 'STORE_SHOW1', 'schema_name': 'TEST3', 'owner_name': 'TEST3',
        # 'is_external': 0, 'hosts': None, 'constraints': None, 'catalog_name': 'CATALOG_SHOW_LIST_STORES'}}]

        # example - json format
        # [{'STORE_INFO': {'STORE_NAME': 'STORE_SHOW1', 'SCHEMA_NAME': 'TEST3', 'OWNER_NAME': 'TEST3',
        # 'IS_EXTERNAL': 0, 'HOSTS': None, 'CONSTRAINTS': None, 'CATALOG_NAME': 'CATALOG_SHOW_LIST_STORES'}}]

        if header['format'] == 'json':
            return reply

        store_info_key = 'store_info'
        catalog_name_key = 'catalog_name'
        store_name_key = 'store_name'
        schema_name_key = 'schema_name'
        hosts_key = 'hosts'
        keyobject_definition_key = 'keyobject_definition'
        keyobject_name_key = 'keyobject_name'
        constraints_name_key = 'constraints'

        info  = res[0][store_info_key]
        catalog_name = info[catalog_name_key]
        store_name = info[store_name_key]
        schema_name = info[schema_name_key]
        owner_name = info["owner_name"]

        constraints_names = info[constraints_name_key]

        if info[hosts_key]:
            tmp = info[hosts_key]
            if type(tmp) is list:
                hosts = [hh[keyobject_definition_key][keyobject_name_key] for hh in tmp]
            elif type(tmp) is dict:
                hosts = [tmp[keyobject_definition_key][keyobject_name_key]]
        else:
            hosts = []

        #return store_name, catalog_name, schema_name, owner_name, hosts, constraints_names
        store = KeySqlStore(store_name, catalog_name, schema_name, owner_name, hosts, constraints_names)
        return store


class KeySqlShowStoresStatement(KeySqlShowStatement):
    """Show Stores Statement  - returns full descriptions of stores
    String query equivalent:
    'show stores for catalog <name>'
    """
    _catalog = None
    _schema = None

    def __init__(self, catalog_name: str = None, schema_name: str = None) -> List[str]:
        """
        :param catalog_name [str]: catalog name for which list of stores to show
        :param schema_name [str]: schema name for which list of stores to show
        """
        self._catalog = catalog_name
        self._schema = schema_name

    def get_catalog_name(self) -> str:
        """
        catalog name
        :return: str - catalog name
        """
        return self._catalog

    def get_schema_name(self) -> str:
        """
        catalog name
        :return: str - schema name
        """
        return self._schema

    def create_request(self) -> str:
        if self._catalog is None and self._schema is None:
            return 'show stores'
        elif self._schema is None:
            return 'show stores for catalog {}'.format(self._catalog)
        elif self._catalog is None:
            return 'show stores from schema {}'.format(self._schema)
        else:
            return 'show stores for catalog {}.{}'.format(self._schema, self._catalog)


    def process_reply(self, reply, header = None) -> list:
        if len(reply) == 0:
            return []

        if header and header['format'] == 'json':
            if type(reply) is str:
                all_items = json.loads(reply)
            else:
                all_items = reply
            #all_items_names = [rec["CATALOG_NAME"] for rec in all_items]
            #return all_items_names
            return all_items

        elif header and header['format'] == 'keysql':
            # list catalogs
            if reply[0].instance_name().lower() == "store_name":
                all_items_names = [rec.instance_value() for rec in reply]
                return all_items_names
            else:   # if reply[0].instance_name().lower() == "store_info":
                # show stores
                all_stores = [KeySqlShowStoreStatement().process_reply([rec], header) for rec in reply]
                return all_stores
        else:
            raise KeySqlException(KeySqlExceptionMessage.FORMAT_KEYSQL_JSON)



class KeySqlShowCatalogStatement(KeySqlShowStatement):
    _name = None

    def __init__(self, name: str = None, schema: str = None):
        self._name = name
        self._schema = schema


    def get_catalog_name(self) -> str:
        """
        catalog name
        :return: str - catalog name
        """
        return self._catalog


    def get_schema_name(self) -> str:
        """
        catalog name
        :return: str - schema name
        """
        return self._schema


    def create_request(self) -> str:
        if self._schema is None:
            return 'show catalog {}'.format(self._name)
        else:
            return 'show catalog {}.{}'.format(self._schema, self._name)


    def process_reply(self, reply, header=None) -> list:
        if reply == "":
            return None # []

        if header and header['format'].lower() not in ['json', 'keysql']:
            raise KeySqlException(KeySqlExceptionMessage.FORMAT_KEYSQL_JSON)

        # """
        if header and header['format'] == 'json':
            if type(reply) is str:
                all_items = json.loads(reply)
            else:
                all_items = reply
            # example of json format result
            # "{"CATALOG_DESCRIPTION":{"SCHEMA_NAME":"TEST3","CATALOG_NAME":"CATALOG02","STORE_NAMES":null,"OWNER_NAME":"test3","KEYOBJECT_DEFINITIONS":null}}"
            # {'CATALOG_DESCRIPTION': {'SCHEMA_NAME': 'TEST3', 'CATALOG_NAME': 'CATALOG_TWIN', 'STORE_NAMES': None, 'OWNER_NAME': 'test3',
            # 'KEYOBJECT_DEFINITIONS': [{'KEYOBJECT_DEFINITION': {'KEYOBJECT_NAME': 'KO1', 'KEYOBJECT_TYPE': 'NUMBER', 'COMPONENT_KEYOBJECT_NAMES': None}}]}}
            return all_items

        # keysql format EXAMPLE:
        """
        # dict looks like
        {CATALOG_DESCRIPTION: {SCHEMA_NAME: 'TEST3', CATALOG_NAME: 'UNIVERSE',
                               STORE_NAMES: {STORE_NAME: 'BILLIARD_STORE'}, OWNER_NAME: 'test3',
                               KEYOBJECT_DEFINITIONS: {
                                   KEYOBJECT_DEFINITION: {KEYOBJECT_NAME: 'MAX_STR', KEYOBJECT_TYPE: 'CHAR',
                                                          COMPONENT_KEYOBJECT_NAMES: NULL},
                                   KEYOBJECT_DEFINITION: {KEYOBJECT_NAME: 'CNT', KEYOBJECT_TYPE: 'NUMBER',
                                                          COMPONENT_KEYOBJECT_NAMES: NULL},
                                   KEYOBJECT_DEFINITION: {KEYOBJECT_NAME: 'AGGR_RESULT', KEYOBJECT_TYPE: 'COMPOSITION',
                                                          COMPONENT_KEYOBJECT_NAMES: {KEYOBJECT_NAME: 'CNT',
                                                                                      KEYOBJECT_NAME: 'MAX_STR'}},
        """

        if header and header['format'] == 'keysql':
            assert reply[0]._name.lower() == 'catalog_info'
            catalog_info = reply[0]._value
            catalog_name = catalog_info['catalog_name']
            schema_name = catalog_info['schema_name']
            owner_name = catalog_info['owner_name']

            if type(catalog_info['store_names']) is list:
                store_names = [rec["store_name"] for rec in catalog_info['store_names']]
            elif type(catalog_info['store_names']) is dict:
                store_names = [catalog_info['store_names']["store_name"]]
            else:
                store_names = []

            ###tmp = KeySqlShowKeyobjectsStatement

            #keyobject_definitions = catalog_info['keyobject_definitions']
            cat_defs = catalog_info['keyobject_definitions']

            if type(cat_defs) is dict:
                dd = ShowKeyObjectStatement.process_keyobject_definition(cat_defs['keyobject_definition'], OrderedDict())
                keyobject_definitions = [val for k, val in list(dd.items())]

            elif type(cat_defs) is list:
                dd = OrderedDict()
                for keyobject_def in cat_defs:
                    dd = ShowKeyObjectStatement.process_keyobject_definition(keyobject_def['keyobject_definition'], dd)
                keyobject_definitions = [val for k, val in list(dd.items())]
            else:
                keyobject_definitions = []

            sequence_definitions = catalog_info['sequence_definitions']

            return KeySqlCatalog(catalog_name, schema_name, owner_name, store_names, keyobject_definitions, sequence_definitions)



class KeySqlShowCatalogsStatement(KeySqlShowStatement):
    # members
    def __init__(self, schema=None, all=False):
        """
        If schema is None, use default schema
        :param schema: if schema is not None
        :param all: if True show all catalogs (full name): show catalogs ALL
                    if False and schema is given show catalogs from schema: show catalogs from [schema] <shcema_name>
        """
        self.schema = schema
        self.all = all

    def create_request(self) -> str:
        if self.all:
            # return 'show catalogs *'
            return 'show catalogs ALL'
        if self.schema:
            # return 'show catalogs {}.*'.format(self.schema)
            # "list catalogs from [schema] <schema_name>"
            return 'show catalogs from schema {}'.format(
                self.schema)  # return 'list catalogs {}.*'.format(self.schema)  #

        if self.schema is None:
            # return 'show catalogs'
            return 'show catalogs'

    def process_reply(self, reply, header=None) -> list:
        if len(reply) == 0: # reply == "":
            return []

        if header and header['format'] == 'json':
            if type(reply) is str:
                all_items = json.loads(reply)
            else:
                all_items = reply
            # all_items_names = [rec["CATALOG_NAME"] for rec in all_items]
            # return all_items_names
            return all_items

        elif header and header['format'] == 'keysql':
            # list catalogs
            if reply[0].instance_name().lower() == "catalog_name":
                all_items_names = [rec.instance_value() for rec in reply]
                return all_items_names
            else:   # if reply[0].instance_name().lower() == "catalog_info":
                # show catalogs
                all_catalogs = [KeySqlShowCatalogStatement().process_reply([rec], header) for rec in reply]
                return all_catalogs
        else:
            raise KeySqlException(KeySqlExceptionMessage.FORMAT_KEYSQL_JSON)

        # remove first { and last}  Catalog <name>{...} and return what is between {}
        # ind = all_items_names.find('{')
        # all_items_names = all_items_names[ind + 1 :-1].split(',\n')
        # if all_items_names[0].upper().startswith('CATALOG'):
        #    all_items_names = all_items_names[1:]
        # return all_items_names.split(',\n')
        # return all_items_names


"""
import json
from collections import namedtuple
from json import JSONEncoder

def customKeySqlCatalogDecoder(catalogDict):
    return namedtuple('CATALOG', catalogDict.keys())(*catalogDict.values())

# dict looks like
# {"CATALOG_DESCRIPTION":{"SCHEMA_NAME":"TEST3","CATALOG_NAME":"CATALOG02","STORE_NAMES":null,"OWNER_NAME":"test3","KEYOBJECT_DEFINITIONS":null}}

CATALOG_NAME
"""


class KeySqlShowCatalogStatement(KeySqlShowStatement):
    _name = None

    def __init__(self, name: str = None):
        self._name = name

    def create_request(self):
        return 'show catalog {}'.format(self._name)

    def process_reply(self, reply, header=None) -> list:
        if reply == "":
            return None # []

        if header and header['format'].lower() not in ['json', 'keysql']:
            raise KeySqlException(KeySqlExceptionMessage.FORMAT_KEYSQL_JSON)

        # """
        if header and header['format'] == 'json':
            if type(reply) is str:
                all_items = json.loads(reply)
            else:
                all_items = reply
            # example of json format result
            # "{"CATALOG_DESCRIPTION":{"SCHEMA_NAME":"TEST3","CATALOG_NAME":"CATALOG02","STORE_NAMES":null,"OWNER_NAME":"test3","KEYOBJECT_DEFINITIONS":null}}"
            # {'CATALOG_DESCRIPTION': {'SCHEMA_NAME': 'TEST3', 'CATALOG_NAME': 'CATALOG_TWIN', 'STORE_NAMES': None, 'OWNER_NAME': 'test3',
            # 'KEYOBJECT_DEFINITIONS': [{'KEYOBJECT_DEFINITION': {'KEYOBJECT_NAME': 'KO1', 'KEYOBJECT_TYPE': 'NUMBER', 'COMPONENT_KEYOBJECT_NAMES': None}}]}}
            return all_items

        # keysql format EXAMPLE:
        """
        # dict looks like
        {CATALOG_DESCRIPTION: {SCHEMA_NAME: 'TEST3', CATALOG_NAME: 'UNIVERSE',
                               STORE_NAMES: {STORE_NAME: 'BILLIARD_STORE'}, OWNER_NAME: 'test3',
                               KEYOBJECT_DEFINITIONS: {
                                   KEYOBJECT_DEFINITION: {KEYOBJECT_NAME: 'MAX_STR', KEYOBJECT_TYPE: 'CHAR',
                                                          COMPONENT_KEYOBJECT_NAMES: NULL},
                                   KEYOBJECT_DEFINITION: {KEYOBJECT_NAME: 'CNT', KEYOBJECT_TYPE: 'NUMBER',
                                                          COMPONENT_KEYOBJECT_NAMES: NULL},
                                   KEYOBJECT_DEFINITION: {KEYOBJECT_NAME: 'AGGR_RESULT', KEYOBJECT_TYPE: 'COMPOSITION',
                                                          COMPONENT_KEYOBJECT_NAMES: {KEYOBJECT_NAME: 'CNT',
                                                                                      KEYOBJECT_NAME: 'MAX_STR'}},
        """

        if header and header['format'] == 'keysql':
            assert reply[0]._name.lower() == 'catalog_info'
            catalog_info = reply[0]._value
            catalog_name = catalog_info['catalog_name']
            schema_name = catalog_info['schema_name']
            owner_name = catalog_info['owner_name']

            if type(catalog_info['store_names']) is list:
                store_names = [rec["store_name"] for rec in catalog_info['store_names']]
            elif type(catalog_info['store_names']) is dict:
                store_names = [catalog_info['store_names']["store_name"]]
            else:
                store_names = []

            ###tmp = KeySqlShowKeyobjectsStatement

            #keyobject_definitions = catalog_info['keyobject_definitions']
            cat_defs = catalog_info['keyobject_definitions']

            if type(cat_defs) is dict:
                dd = ShowKeyObjectStatement.process_keyobject_definition(cat_defs['keyobject_definition'], OrderedDict())
                keyobject_definitions = [val for k, val in list(dd.items())]

            elif type(cat_defs) is list:
                dd = OrderedDict()
                for keyobject_def in cat_defs:
                    dd = ShowKeyObjectStatement.process_keyobject_definition(keyobject_def['keyobject_definition'], dd)
                keyobject_definitions = [val for k, val in list(dd.items())]
            else:
                keyobject_definitions = []

            sequence_definitions = catalog_info['sequence_definitions']

            return KeySqlCatalog(catalog_name, schema_name, owner_name, store_names, keyobject_definitions, sequence_definitions)
