#/**********************************************************************************
#* Copyright (C) 2017-2020 by KEYARK, INC. - All Rights Reserved.
#* All information contained herein is proprietary and confidential to KEYARK, INC.
#* Any use, reproduction, or disclosure without the written
#* permission of KEYARK, INC. is prohibited.
#* Written by Natalia Belyaeva <nbelyaeva@keyark.com>
#***********************************************************************************/
import logging
from typing import List, Optional
from  keysql_client.keysql_statement import KeySqlStatement
from keysql_client.keysql_statement_default import KeySqlDefaultStatement
from keysql_client.keysql_exception import KeySqlException, KeySqlExceptionMessage
from keysql_client.keysql_parser import KeySqlParser
import json

class KeySqlCreateStoreStatement(KeySqlStatement):
    """Create Store Statement
    String query equivalent:
    'create store <store_name> for catalog <catalog_name>'
    """
    _store_name = None
    _catalog_name = None
    _owner_name = None

    def __init__(self, store_name: str=None, catalog_name: str=None, owner: str=None) -> None:
    #def __init__(self, store_name: str, catalog_name: str, schema_name: str="", owner_name: str= "", host_names: list=[], hosts = [], component_names = [], constraints = []) -> None:
        self._store_name = store_name
        self._catalog_name = catalog_name

        # TODO: here?
        """
        self.__name__ = name
        self._catalog_name = catalog_name
        self._owner_name = owner_name
        self._schema_name = schema_name
        self._host_names = host_names if host_names else []
        self._hosts = hosts if hosts else []
        self._comp_names = component_names if component_names else []
        self._constraints = constraints if constraints else []
        """

    def create_request(self) -> str:
        return 'create store {} for catalog {}'.format(self._store_name, self._catalog_name)


class KeySqlDropStoreStatement(KeySqlStatement):
    """
    Drop Store Statement
    String query equivalent:
    'drop store <store_name>'
    """

    def __init__(self, name: str=None) -> None:
        """
        :param name (str): store name to drop
        Example:
       KeySqlDropStoreStatement('store_name')
        """
        self._name = name

    def create_request(self) -> str:
        return 'drop store {}'.format(self._name)


class KeySqlDropStoreIfExistsStatement(KeySqlStatement):
    """
    Drop Store Statement
    String query equivalent:
    'drop store if exists <store_name>'
    """

    def __init__(self, name: str=None) -> None:
        """
        :param name (str): store name to drop
        Example:
       KeySqlDropStoreStatement('store_name')
        """
        self._name = name

    def create_request(self) -> str:
        return 'drop store if exists {}'.format(self._name)


class KeySqlDropStoresStatement(KeySqlStatement):
    """
    Drop Stores Statement
    String query equivalent:
    'drop stores from catalog <catalog_full_name>'
    """

    def __init__(self, catalog_name: str=None) -> None:
        """
        :param name (str): catalog name where drpo all stores
        Example:
       KeySqlDropStoresStatement('catalog_name')
        """
        self._name = catalog_name

    def create_request(self) -> str:
        return 'drop stores {}'.format(self._name)


class KeySqlRenameStoreStatement(KeySqlStatement):
    _old_name = None
    _new_name = None

    def __init__(self, old_name: str = None, new_name: str = None, schema=None):
        self._old_name = old_name
        self._new_name = new_name
        self._schema = schema

    def create_request(self) -> str:
        if self._schema is None:
            return 'rename store {} to {} '.format(self._old_name, self._new_name)
        return 'rename store {} to {} in schema {}'.format(self._old_name, self._new_name,  self._schema)

    def process_reply(self, reply: str, header=None) -> None:
        pass


class KeySqlAlterStoreNameStatement(KeySqlStatement):
    """
    String query equivalent:
    'alter store <store_name> = <new_name> for catalog <catalog_name>
    """
    def __init__(self, store_name: str=None, new_store_name: str=None, catalog_name: str = None) -> None:
        """
        Alter StoreName Statement
        :param store_name (str): old store name
        :param new_store_name (str): new store name, non-empty, unique
        :param catalog_name (str): catalog name
        """
        self._store_name = store_name
        self._new_store_name = new_store_name
        self._catalog_name = catalog_name

    def create_request(self) -> str:
        return 'alter store {} = {} for catalog {}'.format(self._store_name, self._new_store_name, self._catalog_name)

class KeySqlTruncateStoreStatement(KeySqlStatement):
    """
    String query equivalent:
    'Truncate STORE <schema_name.store_name>
    """
    def __init__(self,  schema_name: str = None, store_name: str = None) -> None:
        """
        Truncate StoreName Statement
        :param schema_name (str): schema name
        :param store_name (str): store name, non-empty, unique
        """
        self._schema_name = schema_name
        self._store_name = store_name

    def create_request(self) -> str:
        return "TRUNCATE STORE {}.{}".format(self._schema_name, self._store_name)


class KeySqlDescribeStoreStatement(KeySqlStatement):
    """
    String query equivalent:
    describe store <store_name>
    """
    _store = None
    #_catalog = None

    #def __init__(self, store_name: str = None, catalog_name: str = None) -> None:
    def __init__(self, store_name: str=None) -> None:
        """DescribeStoreStatement
        :param store_name (str): store name
        """
        self._store = store_name
        #self._catalog = catalog_name

    def get_store_name(self) -> str:
        """
        store name
        :return: str - Keyobject name
        """
        return self._store

    def create_request(self) -> str:
        #return 'show store {} for catalog {}'.format(self._store, self._catalog) # as it is uniq, catalog is not used
        return 'describe store {}'.format(self._store)


class KeySqlDescribeStoresStatement(KeySqlStatement):
    """Describe Stores Statement  - returns full descriptions of stores
    String query equivalent:
    'describe stores for catalog <name>'
    """
    _catalog = None

    def __init__(self, catalog_name: str = None) -> List[str]:
        """
        :param catalog_name [str]: Catalog name to describe stores from to show
        """
        self._catalog = catalog_name

    def create_request(self) -> str:
        if self._catalog is None:
            return 'show stores'
        return 'describe stores for catalog {}'.format(self._catalog)


class KeySqlListStoresStatement(KeySqlStatement):
    """List Stores Statement - returns list of store names
    String query equivalent:
    'list stores for catalog <name>'
    """
    _catalog = None

    def __init__(self, catalog_name: str = None) -> List[str]:
        """
        :param catalog_name [str]: List of stores to show
        """
        self._catalog = catalog_name

    def create_request(self) -> str:
        if self._catalog is None:
            return 'list stores'
        return 'list stores for catalog {}'.format(self._catalog)

    def process_reply(self, reply:str, header=None) -> List:
        if reply == '' or reply == []:
            return []

        # reply: like  [{"STORE_NAME":"DUMMY_STORE1"},{"STORE_NAME":"DUMMY_STORE2"}]
        if header and header['format'] == 'json':
            if type(reply) is list:
                return [rec["STORE_NAME"] for rec in reply]

        if header and header['format'] == 'keysql':
            if type(reply) is list:
                #print(reply, '--------------------------------------------err check')
                #print(reply[0], '--------------------------------------------err check')
                #print(reply[0].instance_name(), '--------------------------------------------err check')
                if reply[0].instance_name() != 'STORE_NAME':
                    raise KeySqlException(KeySqlExceptionMessage.INSTANCE_KEY_SCHEMA_NAME)
                return [rec.instance_value() for rec in reply]

        #TODO:
        raise Exception("_______ Not IMPLEMENTED keysql_store_statement _____________________")
        # else: 'keysql' format
        #TODO: remove as obsolete
        if type(reply) is str:
            all_items_names = reply.strip()
            all_items_names = all_items_names.split(',\n')
        return all_items_names


if __name__ == "__main__":

    create_store = KeySqlCreateStoreStatement('store_name')
    drop_store = KeySqlDropStoreStatement('store_name')
    drop_stores = KeySqlShowStoresStatement(['store1', 'store2'])

    # catalog
    # {CATALOG_DESCRIPTION:{SCHEMA_NAME:'TEST3',CATALOG_NAME:'XYZ',STORE_NAMES:NULL,
    # OWNER_NAME:'test3',
    # KEYOBJECT_DEFINITIONS:{KEYOBJECT_DEFINITION:{KEYOBJECT_NAME:'NN',KEYOBJECT_TYPE:'NUMBER',COMPONENT_KEYOBJECT_NAMES:NULL},
    # KEYOBJECT_DEFINITION:{KEYOBJECT_NAME:'CC',KEYOBJECT_TYPE:'CHAR',COMPONENT_KEYOBJECT_NAMES:NULL},
    # KEYOBJECT_DEFINITION:{KEYOBJECT_NAME:'CMPND',KEYOBJECT_TYPE:'COMPOSITION',COMPONENT_KEYOBJECT_NAMES:{KEYOBJECT_NAME:'NN',KEYOBJECT_NAME:'CC'}}}}}
    # store
    # {STORE_DESCRIPTION:{SCHEMA_NAME:'TEST3',CATALOG_NAME:'PEOPLE',STORE_NAME:'PEOPLE_STORE',
    # OWNER_NAME:'test3',
    # HOSTS:{KEYOBJECT_DEFINITION:{KEYOBJECT_NAME:'PERSON',KEYOBJECT_TYPE:'COMPOSITION',
    # COMPONENT_KEYOBJECT_NAMES:{KEYOBJECT_NAME:'EMAILS',KEYOBJECT_NAME:'FIRST_NAME',KEYOBJECT_NAME:'LAST_NAME',KEYOBJECT_NAME:'YOB',KEYOBJECT_NAME:'ADDRESS'}}},
    # CONSTRAINTS:NULL}}


    store_keysql = "{STORE_DESCRIPTION:{SCHEMA_NAME:'TEST3',CATALOG_NAME:'PEOPLE',STORE_NAME:'PEOPLE_STORE',OWNER_NAME:'test3',HOSTS:{KEYOBJECT_DEFINITION:{KEYOBJECT_NAME:'PERSON',KEYOBJECT_TYPE:'COMPOSITION',COMPONENT_KEYOBJECT_NAMES:{KEYOBJECT_NAME:'EMAILS',KEYOBJECT_NAME:'FIRST_NAME',KEYOBJECT_NAME:'LAST_NAME',KEYOBJECT_NAME:'YOB',KEYOBJECT_NAME:'ADDRESS'}}},CONSTRAINTS:NULL}}"

    reply1 = KeySqlParser.parse_select_reply(store_keysql)

    print(reply1)