"""
#/**********************************************************************************
#* Copyright (C) 2017-2020 by KEYARK, INC. - All Rights Reserved.
#* All information contained herein is proprietary and confidential to KEYARK, INC.
#* Any use, reproduction, or disclosure without the written
#* permission of KEYARK, INC. is prohibited.
#* Written by Natalia Belyaeva <nbelyaeva@keyark.com>
#***********************************************************************************/
"""

import logging
from keysql_client.keysql_exception import KeySqlExceptionMessage
from keysql_client.keysql_connector import *
from keysql_client.keysql_statement import KeySqlStatement

class UpdateStatement(KeySqlStatement):
    """Update Statement
    Example:
    statement = UpdateStatement("billiard_store").set(my_brand).set(my_diam).add("balls",my_new_ball).remove("balls",my_old_ball).where(":num is null and :barnd = 'Carambo'")
    """
    _store_name = None
    _where = None
    _set = {}
    _add = []
    _remove = []

    def __init__(self, store_name: str = None):
        """Create UpdateStatement
        :param store_name (str): store name
        """
        self._store_name = store_name
        self._where = None
        self._set = {}
        self._add = []
        self._remove = []

    def set(self, instance: dict):
        """
        :param instance (dict): dictionary of update instances
        :return: self
        Example:
        count = conn.execute(store.update().where(":last_name = 'Johns'").set({"address" : {"country" : "Russia"}}))
        """
        self._set = instance
        return self

    def __add__(self, name, instance: str):
        """Add instance
        :param name (str): instance name
        :param instance (str):
        :return: self
        """
        self._add.append((name, instance))
        return self

    def __remove__(self, name, instance: str):
        """Remove instance
        :param name (str): instance name
        :param instance (str):
        :return: self
        """
        self._remove.append((name, instance))
        return self

    def where(self, where_str:str):   # ? TODO: check later: where(self, where_str:str, *args)
        """Where Clause
        :param where_str (str): condition as a string
        :return:
        Example:
        count = conn.execute(store.update().where(":last_name = 'Johns'").set({"address.city": "Moscow"}))
        """
        self._where = where_str
        return self

    def create_request(self) -> str :
        if self._store_name is None:
            raise KeySqlException(KeySqlExceptionMessage.STORE_IS_NOT_DEFINED)
        request = "update " + self._store_name
        first = True
        if len(self._set.keys()) > 0:
            request += " set "
            items = []
            for key in self._set.keys():
                value = self._set[key]
                if type(value) == str:
                    value = "'" + value + "'"
                elif type(value) == dict:
                    val = "{"
                    for k, v in value.items():
                        val += k
                        val += ":"
                        val += "'" + v + "'"
                    value = val + "}"
                else:
                    value = str(value)
                items.append(key + "=" + value)
            request += ', '.join(items)

        for x in self._add:
            if first:
                first = False
            else:
                request += ", "
                request += x[0] + " += " + x[1].serialize()
        for x in self._remove:
            if first:
                first = False
            else:
                request += ", "
                request += x[0] + " += " + x[1].serialize()

        if self.where:
            request += " where " + self._where

        logging.debug('keysql_store_update request: ' + request)
        return request

    def process_reply(self, reply: str, header=None):
        count = header['instance_count'] if header else 0
        return count